<?php

namespace SGalinski\SgYoutube\Service;

use InvalidArgumentException;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

/**
 * YouTube Helper Service
 */
class YoutubeService {
	const API_URL = 'https://www.googleapis.com/youtube/v3/';
	const API_CHANNEL = 'search';
	const API_PLAYLIST = 'playlistItems';
	const API_VIDEO = 'videos';
	const API_PART = 'snippet';
	const API_ORDER_BY = 'date';

	/**
	 * Maps the json array from the YouTube call to return some unified value. The output from YouTube is pretty
	 * unsteady. Also we calculate the correct thumbnail sized and so on.
	 *
	 * @param array $jsonArray
	 * @param string $youtubeId
	 * @param string $aspectRatio 16:9 (default) or 4:3 (ONLY used if byAspectRation is set as thumbnail type)
	 * @param string $thumbnailType maxres, standard, high, medium, default, byAspectRatio (default)
	 * @return array
	 */
	public function mapArray(
		$jsonArray = [], $youtubeId = '', $aspectRatio = '16:9', $thumbnailType = 'byAspectRatio'
	): array {
		if (!in_array($thumbnailType, ['maxres', 'standard', 'high', 'medium', 'default', 'byAspectRatio'])) {
			$thumbnailType = 'byAspectRatio';
		}

		if (!in_array($aspectRatio, ['16:9', '4:3'])) {
			$aspectRatio = '16:9';
		}

		$result = [];
		foreach ($jsonArray as $field) {
			if (strpos($youtubeId, 'UC') === 0) {
				$youTubeIdFromArray = $field['id']['videoId'];

				// prevent the channel preview
				if ($field['id']['kind'] !== 'youtube#video') {
					continue;
				}

			} elseif (strpos($youtubeId, 'PL') === 0) {
				$youTubeIdFromArray = $field['snippet']['resourceId']['videoId'];
			} else {
				$youTubeIdFromArray = $field['id'];
			}

			$previewImage = [];
			$resolutionTypes = ['maxres', 'standard', 'high', 'medium', 'default'];
			if ($thumbnailType !== 'byAspectRatio') {
				array_unshift($resolutionTypes, $thumbnailType);
			}

			foreach ($resolutionTypes as $type) {
				if (isset($field['snippet']['thumbnails'][$type])) {
					$previewImage = $field['snippet']['thumbnails'][$type];
					if ($thumbnailType === 'byAspectRatio' && isset($previewImage['height'])) {
						$aspectRatioOfImage = $previewImage['width'] / $previewImage['height'];
						if ($aspectRatio === '16:9' && $aspectRatioOfImage > 1.7 && $aspectRatioOfImage < 1.9) {
							break;
						}

						if ($aspectRatio === '4:3' && $aspectRatioOfImage > 1.2 && $aspectRatioOfImage < 1.4) {
							break;
						}
					} else {
						break;
					}
				}
			}

			// Don't cache the preview URL. YouTube has a CDN and delivers much faster.
			$result[] = [
				'title' => $field['snippet']['title'],
				'description' => strip_tags($field['snippet']['description']),
				'thumbnail' => $previewImage['url'],
				'url' => 'https://www.youtube.com/watch?v=' . $youTubeIdFromArray
			];
		}

		return $result;
	}

	/**
	 * Returns a JSON array with the video details (title, description, preview image, url)
	 *
	 * @param string $youtubeId
	 * @param string $maxResults
	 * @param string $apiKey
	 * @return array|mixed
	 */
	public function getJsonAsArray($youtubeId = '', $maxResults = '10', $apiKey = '') {
		$url = $this->getApiUrl($youtubeId, $maxResults, $apiKey);

		if (function_exists('curl_init')) {
			$ch = curl_init();
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, TRUE);
			curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
			curl_setopt($ch, CURLOPT_URL, $url);
			$jsonString = curl_exec($ch);
			curl_close($ch);

		} elseif (ini_get('allow_url_fopen')) {
			$jsonString = file_get_contents($url);
		} else {
			throw new InvalidArgumentException(
				'The system doesn\'t have CURL enabled and "allow_url_fopen" is disabled.', 403
			);
		}

		$jsonArray = ($jsonString !== '' ? json_decode($jsonString, TRUE) : []);
		if ($jsonArray === NULL) {
			throw new InvalidArgumentException(
				'There is something wrong with loaded JSON or encoded data is deeper than the recursion limit.', 403
			);
		}

		if (!$jsonArray) {
			throw new InvalidArgumentException('JSON could\'t be parsed.', 403);
		}

		// By API documentation provided 'error' key is sent if, probably,
		// URL can not return JSON data or Permission is denied.
		if (isset($jsonArray['error'])) {
			throw new InvalidArgumentException('Message: ' . $jsonArray['error']['message'], 403);
		}

		if (!isset($jsonArray['items'])) {
			throw new InvalidArgumentException('No items array.', 403);
		}

		if (count($jsonArray['items']) < 1) {
			throw new InvalidArgumentException('No items found.', 403);
		}

		return $jsonArray;
	}

	/**
	 * Returns the YouTube API URL
	 *
	 * @param string $youtubeId
	 * @param string $maxResults
	 * @param string $apiKey
	 * @return string
	 */
	public function getApiUrl($youtubeId = '', $maxResults = '10', $apiKey = ''): string {
		$apiUrl = self::API_URL;
		$parameters = [];

		if (strpos($youtubeId, 'UC') === 0) {
			$apiUrl .= self::API_CHANNEL;
			$parameters['channelId'] = $youtubeId;
		} elseif (strpos($youtubeId, 'PL') === 0) {
			$apiUrl .= self::API_PLAYLIST;
			$parameters['playlistId'] = $youtubeId;
		} else {
			$apiUrl .= self::API_VIDEO;
			$parameters['id'] = $this->removeIdParameters($youtubeId);
		}

		$parameters['order'] = self::API_ORDER_BY;
		$parameters['part'] = self::API_PART;
		$parameters['key'] = $apiKey;
		$parameters['maxResults'] = $maxResults;
		$query = http_build_query($parameters);

		return $apiUrl . '?' . $query;
	}

	/**
	 * Removes GET parameters following the ID
	 *
	 * @param string $youtubeId
	 * @return array|string
	 */
	protected function removeIdParameters($youtubeId = '') {
		if (strpos($youtubeId, '&')) {
			return explode('&', $youtubeId)[0];
		}

		return $youtubeId;
	}
}
